import unittest
import importlib.resources
import json

from typing import Optional

from deephaven.plugin.object import Exporter, Reference, find_object_type


def read_text(resource: str):
    return importlib.resources.read_text(__package__, resource)


def read_json(resource: str):
    return json.loads(read_text(resource))


class TestExporter(Exporter):
    def __init__(self):
        self._objects = []
        self._refs = []

    def reference(self,
                  object,
                  allow_unknown_type: bool = False,
                  force_new: bool = False) -> Optional[Reference]:
        if not force_new:
            ix = 0
            for o in self._objects:
                if object is o:
                    return self._refs[ix]
                ix += 1
        object_type = find_object_type(object)
        if not object_type and not allow_unknown_type:
            return None
        ref = Reference(
            len(self._refs),
            object_type.name if object_type else None,
            b'Not-Real-Server')
        self._objects.append(object)
        self._refs.append(ref)
        return ref


class TestJsonPluginLocal(unittest.TestCase):
    def test_plain_python_types(self):
        from deephaven.plugin.json import Encoder
        from .plain_python_types import output
        # Exporter not needed for simple types
        actual_json = json.loads(Encoder(None).encode(output))
        expected_json = read_json('plain_python_types.json')
        self.assertEqual(actual_json, expected_json)

    def test_nested_node_types(self):
        from deephaven.plugin.json import Encoder
        from .nested_node_types import output
        actual_json = json.loads(Encoder(TestExporter()).encode(output))
        expected_json = read_json('nested_node_types.json')
        self.assertEqual(actual_json, expected_json)

    def test_nested_other_types(self):
        from deephaven.plugin.json import Encoder
        from .nested_other_types import output
        actual_json = json.loads(Encoder(TestExporter()).encode(output))
        expected_json = read_json('nested_other_types.json')
        self.assertEqual(actual_json, expected_json)

    def test_unknown_python_types(self):
        from deephaven.plugin.json import Encoder
        from .unknown_python_types import output
        actual_json = json.loads(Encoder(TestExporter()).encode(output))
        expected_json = read_json('unknown_python_types.json')
        self.assertEqual(actual_json, expected_json)

# class TestJsonPluginRemote(unittest.TestCase):
#     def execute_code(self, code: str):
#         # todo: execute code, inspect output is as expected
#         pass

#     def fetch_json_output(self):
#         return json.loads('"TODO"')

#     def check_json(self, name):
#         code = read_text(f'{name}.py')
#         expected_json = read_json(f'{name}.json')
#         self.execute_code(code)
#         output_json = self.fetch_json_output()
#         self.assertEqual(output_json, expected_json)

#     def test_plain_python_types(self):
#         self.check_json('plain_python_types')

#     def test_nested_node_types(self):
#         self.check_json('nested_node_types')

#     def test_nested_other_types(self):
#         self.check_json('nested_other_types')


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import json

from deephaven.plugin import Registration
from deephaven.plugin.object \
    import Exporter, ObjectType, Reference

__version__ = "0.0.1.dev4"


class Node:
    def __init__(self, object, is_ref=False, **kw) -> None:
        self._object = object
        self._is_ref = is_ref
        self._kw = kw

    @property
    def object(self):
        return self._object

    @property
    def kw(self):
        return self._kw

    @property
    def is_ref(self) -> bool:
        return self._is_ref

    @property
    def as_ref(self) -> Node:
        return Node(self._object, is_ref=True, **self._kw)


class Encoder(json.JSONEncoder):
    def __init__(self, exporter: Exporter, **kw):
        super().__init__(**kw)
        self._exporter = exporter

    def default(self, obj):
        if isinstance(obj, Node):
            if obj.is_ref:
                ref = self._exporter.reference(obj)
                if not ref:
                    raise RuntimeError("Unable to create reference to Node")
                return ref
            else:
                return obj.object

        if isinstance(obj, Reference):
            # Note: serializing type is extraneous,
            # but it makes our testing easier. Should
            # we provide a way to turn type serialization on/off?
            return {
                'type': obj.type,
                'index': obj.index}

        # Try to create a reference with a known type
        ref = self._exporter.reference(obj, allow_unknown_type=False)
        if ref:
            return ref

        # Try the default JSON (python) types
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            pass

        # Fallback to unknown reference types
        ref = self._exporter.reference(obj, allow_unknown_type=True)
        if not ref:
            raise RuntimeError("Unable to create reference from exporter")
        return ref


class JsonType(ObjectType):
    @property
    def name(self) -> str:
        return "deephaven.plugin.json.Node"

    def is_type(self, object) -> bool:
        return isinstance(object, Node)

    def to_bytes(self, exporter: Exporter, node: Node) -> bytes:
        encoder_kw = dict(node.kw)
        encoder_cls = encoder_kw.pop('cls', Encoder)
        if issubclass(encoder_cls, Encoder):
            encoder = encoder_cls(exporter, **encoder_kw)
        else:
            # todo: should this be an error case?
            encoder = encoder_cls(**encoder_kw)
        json_text = encoder.encode(node.object)
        return str.encode(json_text)


class JsonRegistration(Registration):
    @classmethod
    def register_into(cls, callback: Registration.Callback) -> None:
        callback.register(JsonType)
